#
# Anbox Stream Agent
# Copyright 2019 Canonical Ltd.  All rights reserved.
#

import os
import subprocess
import socket

import charms.coordinator
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import templating
from charmhelpers.core.hookenv import (log, status_set, config, relation_ids,
                                       application_version_set, is_leader,
                                       leader_get, leader_set, unit_private_ip)
from charmhelpers.core.host import service, service_restart
from charms.layer.anbox import (generate_random_string,
                                install_or_upgrade_anbox_package,
                                remove_anbox_package)
from charms.layer.pki import (create_cert_files_if_needed, read_cert)
from charms.reactive import hook, when, when_not, when_any
from charms.reactive.flags import clear_flag, is_flag_set, set_flag
from charms.reactive.relations import endpoint_from_flag

SERVICE = 'snap.anbox-stream-agent.anbox-stream-agent.service'

SNAP_COMMON_PATH = '/var/snap/anbox-stream-agent/common'
AGENT_PATH = '{}/agent'.format(SNAP_COMMON_PATH)
TLS_DIR = '{}/tls'.format(SNAP_COMMON_PATH)
TLS_PRIVATE_CERT = '{}/private_cert.pem'.format(TLS_DIR)
TLS_PRIVATE_KEY = '{}/private_key.pem'.format(TLS_DIR)
AMS_CERT_BASE_PATH = '{}/ams'.format(SNAP_COMMON_PATH)
AMS_CLIENT_CERT_PATH = '{}/client.crt'.format(AMS_CERT_BASE_PATH)
AMS_CLIENT_KEY_PATH = '{}/client.key'.format(AMS_CERT_BASE_PATH)
AGENT_CONFIG_PATH = '{}/config.yaml'.format(AGENT_PATH)
NATS_PATH = '{}/nats'.format(SNAP_COMMON_PATH)
NATS_CA_PATH = '{}/ca.pem'.format(NATS_PATH)
AGENT_ID_PATH = '{}/id'.format(AGENT_PATH)

DEB_PACKAGE_NAME = "anbox-cloud-stream-agent"
DEFAULT_AGENT_PORT_NUMBER = 8082


class NotFoundException(Exception):
    pass


def set_status_from_endpoints():
    ams = endpoint_from_flag('endpoint.ams.joined')
    nats = endpoint_from_flag('endpoint.nats.joined')
    nats_url = config('nats_url') or ""
    if not ams:
        status_set('maintenance', 'Waiting for AMS')
    elif not nats and not len(nats_url) == 0:
        status_set('maintenance', 'Waiting for NATs')


@when('ua.ready')
@when_not('anbox-stream-agent.installed')
def install():
    install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)
    set_charm_version()
    set_flag('anbox-stream-agent.installed')


@when('certificates.available', 'anbox-stream-agent.installed')
def request_certificates(tls):
    if not os.path.exists(TLS_DIR):
        os.mkdir(TLS_DIR, mode=0o700)

    # Private IP certificate for internal connections
    subject_alt_names = set()
    subject_alt_names.add(socket.gethostname())
    subject_alt_names.add(unit_private_ip())
    subject_alt_names = list(subject_alt_names)
    tls.request_server_cert(unit_private_ip(), subject_alt_names)


@when_any('certificates.certs.changed', 'certificates.certs.available')
@when('anbox-stream-agent.installed')
def update_certs():
    if is_flag_set('anbox-stream-agent.certificates_set'):
        return
    ca = endpoint_from_flag('certificates.certs.available')
    if not ca.server_certs:
        return

    status_set('maintenance', 'Configuring CA')
    log('Certificates: updating certificates', 'INFO')

    place_certs(ca)

    set_flag('anbox-stream-agent.certificates_set')
    configure()
    clear_flag('certificates.certs.changed')


def place_certs(ca):
    """
    place_certs gets all certificates from the CA relation and puts them
    in their appropriate location
    """
    # For each unit, we only have one certificate as we use the private
    # IP as CN to relate with other charms.
    private_ip_cert = None

    for cert in ca.server_certs:
        if cert.common_name == unit_private_ip():
            private_ip_cert = cert
        else:
            log('Certificates: invalid CN ' + cert.common_name, 'ERROR')
            status_set("blocked", "Error setting up certificates")
            return

    if private_ip_cert is None:
        log('Certificates: Empty certificates', 'ERROR')
        status_set("blocked", "Error setting up certificates")
        return

    fmode = os.O_CREAT | os.O_WRONLY
    # Private IP certificate
    with open(os.open(TLS_PRIVATE_KEY, fmode, 0o600), "w+") as f:
        f.write(private_ip_cert.key)
    with open(os.open(TLS_PRIVATE_CERT, fmode, 0o600), "w+") as f:
        f.write(private_ip_cert.cert)

    log('Certificates: Updated certificates', 'INFO')


def set_charm_version():
    try:
        info = subprocess.run(['snap', 'info', 'anbox-stream-agent'],
                              stdout=subprocess.PIPE)
    except subprocess.CalledProcessError as err:
        log('Failed to get snap info: {}'.format(err), level='ERROR')
    else:
        stdout = info.stdout.decode('utf8').splitlines()
        version = [line.split()[1] for line in stdout if 'installed' in line]
        if version:
            application_version_set(version[0])


@when('endpoint.nats.changed', 'anbox-stream-agent.installed', 'nats.configured')
def nats_relation_changed(*args):
    clear_flag('endpoint.nats.changed')
    clear_flag('anbox-stream-agent.configured')


@when('endpoint.nats.departed')
def nats_relation_departed(*args):
    clear_flag('endpoint.nats.departed')
    clear_flag('anbox-stream-agent.configured')


@when('endpoint.ams.changed')
def ams_relation_changed(*args):
    clear_flag('endpoint.ams.changed')
    clear_flag('anbox-stream-agent.configured')


@when('endpoint.ams.departed')
def ams_relation_departed(*args):
    clear_flag('endpoint.ams.departed')
    clear_flag('anbox-stream-agent.configured')
    clear_flag('ams.certs_published')


@when('endpoint.stun.changed')
def stun_changed(*args):
    clear_flag('anbox-stream-agent.configured')
    clear_flag('endpoint.stun.changed')


@when('endpoint.stun.departed')
def stun_departed(*args):
    clear_flag('anbox-stream-agent.configured')
    clear_flag('endpoint.stun.departed')


@hook('upgrade-charm')
def upgrade_charm():
    """Sets the charm as active after upgrading the charm"""
    if not is_flag_set('anbox-stream-agent.installed'):
        return

    install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)

    set_charm_version()
    clear_flag('anbox-stream-agent.configured')

    # If we have an NRPE relation, force run of that as well
    if relation_ids('nrpe-exteral-master'):
        clear_flag('nagios.configured')


@hook('stop')
def stop():
    """Remove the charm main snap when charm is removed"""
    if not is_flag_set('anbox-stream-agent.installed'):
        return

    remove_anbox_package(DEB_PACKAGE_NAME)


def get_ams_config():
    """Returns the first AMS configuration received through relationship"""
    ams = endpoint_from_flag('endpoint.ams.joined')
    if not ams or not ams.services():
        log('AMS relation data still not available. Skip')
        return None

    ams_config = {}
    for srv in ams.services():
        for k, v in srv.items():
            ams_config[k] = v
        break
    return ams_config


@when_any('config.changed.agent_id',
          'config.changed.region',
          'config.changed.nats_ca',
          'config.changed.nats_url',
          'config.changed.nats_urls',
          'config.changed.expose_tcp_port',
          'config.changed.coturn_urls',
          'config.changed.coturn_auth_secret',
          'config.changed.turn_auth_max_lifetime',
          'config.changed.extra_stun_servers',
          'config.changed.location',
          'config.changed.port')
def config_changed():
    clear_flag('anbox-stream-agent.configured')


@when('anbox-stream-agent.installed')
@when_not('anbox-stream-agent.configured')
def configure():
    # Do not configure the agent if certs are not available
    if not is_flag_set('anbox-stream-agent.certificates_set'):
        status_set('blocked', 'Waiting for CA and certificates')
        return

    """
    Configure generates the main configuration for the stream agent.
    It stops early if the required relation are not joined.
    """
    log('Reconfiguring the Agent', level='INFO')
    # We either have a relation with NATs or have the URL and
    # CA certificate directly configured via the charm config
    nats_ca_cert = config('nats_ca') or ""
    nats_urls = []
    urls = config('nats_urls') or ""
    if len(urls) > 0:
        nats_urls += urls.split("\n")
    nats_url = config('nats_url') or ""
    if len(nats_url) > 0:
        nats_urls.append(nats_url)

    # If we have a relation to NATs it takes priority over the
    # configured values
    nats = endpoint_from_flag('endpoint.nats.joined')
    if nats:
        nats_ca_cert = nats.ca_cert() or ""
        nats_urls += nats.urls()

    if len(nats_urls) == 0:
        log("NATs endpoint is not available", level='WARNING')
        status_set('blocked', 'Waiting for NATs')
        service('disable', SERVICE)
        return

    raw_extra_stun_servers = config('extra_stun_servers') or ""
    extra_stun_servers = []
    if len(raw_extra_stun_servers) > 0:
        extra_stun_servers = raw_extra_stun_servers.split(",")

    coturn = {'urls': [], 'auth_secret': ''}
    coturn_urls = config('coturn_urls')
    if len(coturn_urls) > 0:
        coturn['urls'] = coturn_urls.split(',')
        coturn['auth_secret'] = config('coturn_auth_secret')

    if len(coturn['urls']) == 0:
        stun = endpoint_from_flag('endpoint.stun.joined')
        if stun:
            stun_servers = stun.get_servers()
            urls = []
            for server in stun_servers:
                urls.extend(server['urls'])

            # Deduplicate entries in the URL list. In case of multiple coturn units
            # and a location set (as coturn is behind an lb) we need to deduplicate
            # the urls here to avoid duplicates
            cleaned_urls = list(set(urls))

            coturn['urls'] = []
            for url in cleaned_urls:
                # STUN doesn't need authentication so put it into the extra STUN
                # server list we configure the agent with
                if url.startswith("stun:"):
                    extra_stun_servers.extend([url])
                    continue
                coturn['urls'].extend([url])

            # We currently can't set multiple auth secrets in the agent
            # configuration so we rely on the coturn charm to set the
            # same secret for all servers it supplies via the relation
            if len(stun_servers) > 0:
                coturn['auth_secret'] = stun_servers[0]['auth_secret']

    coturn['max_lifetime'] = config('turn_auth_max_lifetime')

    ams = endpoint_from_flag('endpoint.ams.joined')
    if not ams:
        log("AMS endpoint is not available")
        service('disable', SERVICE)
        status_set('blocked', 'Waiting for AMS')
        return
    create_cert_files_if_needed(AMS_CLIENT_CERT_PATH, AMS_CLIENT_KEY_PATH)

    if not is_flag_set('ams.certs_published'):
        cert = read_cert(AMS_CLIENT_CERT_PATH)
        ams.publish_info(client_certificate=cert)
        set_flag('ams.certs_published')

    ams_cfg = get_ams_config()
    ams_url = ""
    if ams_cfg and 'private_address' in ams_cfg and 'port' in ams_cfg:
        ams_url = 'https://{}:{}'.format(ams_cfg['private_address'],
                                         ams_cfg['port'])
    if not ams_cfg or len(ams_url) == 0:
        log("AMS endpoint is not available")
        status_set('blocked', 'Waiting for AMS')
        service('disable', SERVICE)
        return

    if not os.path.exists(AGENT_PATH):
        os.mkdir(AGENT_PATH, mode=0o700)

    # If no id is set we generate one and store it to be able to
    # reuse it later on
    agent_id = config('agent_id')
    if len(agent_id) == 0:
        if not os.path.exists(AGENT_ID_PATH):
            agent_id = generate_random_string(12)
            with open(AGENT_ID_PATH, "w+") as f:
                f.write(agent_id)
        else:
            with open(AGENT_ID_PATH, "r") as f:
                agent_id = f.read()

    region = config('region')
    if len(region) == 0:
        region = 'unknown'

    if nats_ca_cert is not None and len(nats_ca_cert) > 0:
        if not os.path.exists(NATS_PATH):
            os.makedirs(NATS_PATH, mode=0o0700)
        flags = os.O_CREAT | os.O_WRONLY
        with open(os.open(NATS_CA_PATH, flags, 0o600), 'w') as f:
            f.write(nats_ca_cert)

    expose_tcp_port = config("expose_tcp_port") or False

    auth_secret_key = find_or_generate_secret_key()
    if auth_secret_key is None:
        status_set('blocked', 'Failed to generate secret key')
        service('disable', SERVICE)
        return

    port = config('port') or DEFAULT_AGENT_PORT_NUMBER
    templating.render(
        source='config.yaml',
        target=AGENT_CONFIG_PATH,
        context={
            'listen_address': "{}:{}".format(unit_private_ip(), port),
            'location': config('location') or "",
            'tls_key_path': TLS_PRIVATE_KEY,
            'tls_cert_path': TLS_PRIVATE_CERT,
            'auth_secret_key': auth_secret_key,
            'ams_url': ams_url,
            'agent_id': agent_id,
            'region': region,
            'nats_urls': nats_urls,
            'nats_tls_ca_path': NATS_CA_PATH,
            'nats_tls_ca': len(nats_ca_cert) > 0,
            'coturn': coturn,
            'expose_tcp_port': expose_tcp_port,
            'extra_stun_servers': extra_stun_servers,
        },
        perms=0o600)

    set_flag('anbox-stream-agent.configured')
    trigger_service_restart()


def find_or_generate_secret_key():
    if is_leader():
        auth_secret = leader_get('auth_secret_key')
        if not auth_secret:
            auth_secret = generate_random_string(64)
            auth_info = {'auth_secret_key': "{}".format(auth_secret)}
            leader_set(auth_info)
        return auth_secret
    else:
        return leader_get('auth_secret_key')


def trigger_service_restart():
    """
    The restart is controlled by the coordinator layer
    so the agent may not restart immediately
    """
    status_set('maintenance', 'Restarting')
    charms.coordinator.acquire('restart')


@when('coordinator.granted.restart')
def _restart_service():
    """
    Restarts snap main service.
    DO NOT CALL DIRECTLY. To ensure maximum uptime in HA, all units are
    restarted at different times. This order is guaranteed by the coordinator
    layer.
    """
    status_set('maintenance', 'Restart')
    log("Lock acquired, restarting...", level='INFO')
    service('enable', SERVICE)
    service_restart(SERVICE)
    status_set('active', '')
    log("Restarted, releasing lock", level='INFO')


def detect_cpu_arch():
    try:
        output = subprocess.check_output(
            ['dpkg', '--print-architecture']).strip()
        return str(output, 'utf-8')
    except subprocess.CalledProcessError as ex:
        log('Error detecting current cpu architecture: {} - {}'.format(
                ex.returncode, str(ex.output, 'utf-8')))
        raise ex


@when_any('config.changed.nagios_context',
          'config.changed.nagios_servicegroups')
def nagios_config_changed():
    if not is_flag_set('endpoint.nrpe-external-master.joined'):
        return
    update_nrpe_config()


@when('nrpe-external-master.available')
@when_not('nagios.configured')
def update_nrpe_config():
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    arch = detect_cpu_arch()
    cmd = [
        '/usr/lib/nagios/plugins/check_procs',
        '-c', '1:', '-C', 'agent.{}'.format(arch)
    ]
    nrpe_setup.add_check(
        shortname='anbox-stream-agent_proc',
        description='Check Anbox Stream Agent process',
        check_cmd=" ".join(cmd),
    )
    nrpe_setup.write()
    set_flag('nagios.configured')
